open Client

(* ∀a b. ({} -> (a * (a -> b))) -> b
   Λa b. fun (p : {} -> (a * (a -> b))) ->
     let (x, f) = p () in f x
*)
let let_prod =
  let open F in
  let alpha, beta = 0, 1 in
  let p, f, x = "p", "f", "x" in
  TyAbs (dummy_range, alpha, TyAbs (dummy_range, beta,
    Abs (dummy_range, p, TyArrow (TyProduct [],
           TyProduct [TyVar alpha; TyArrow (TyVar alpha, TyVar beta)]),
         LetProd (dummy_range, [x; f],
                  App (dummy_range,
                       Var (dummy_range, p),
                       Tuple (dummy_range, [])),
        App (dummy_range, Var (dummy_range, f), Var (dummy_range, x))))))

(* Λαβ. λ(x:{α*β}. match x with (_, y) -> y end *)
let match_with_prod =
  let alpha, beta = 1, 0 in
  F.(
    TyAbs(dummy_range, alpha,
      TyAbs(dummy_range, beta,
        Abs(dummy_range, "x",
          TyProduct [ TyVar alpha ; TyVar beta ],
          Match(dummy_range, TyVar beta,
                Var (dummy_range, "x"),
                [
                  (PTuple (dummy_range,
                           [ PWildcard dummy_range ; PVar (dummy_range, "y") ]) ,
                   Var (dummy_range, "y"))
                ]
            )
          )
        )
      )
  )

(* option *)
let option_env =
  let option_typedecl =
    let open Datatype in {
        name = Type "option";
        type_params = [ 0 ];
        data_kind = Variant;
        labels_descr = [
          {
            label_name = Label "None";
            type_name = Type "option";
            arg_type = F.TyProduct [];
          } ; {
            label_name = Label "Some";
            type_name = Type "option";
            arg_type = F.TyVar 0;
          }
        ];
      }
  in
  Datatype.Env.add_decl Datatype.Env.empty option_typedecl

let unit = F.Tuple (F.dummy_range, [])
let unit_pattern = F.PTuple (F.dummy_range, [])

(* None *)
let none =
  let alpha = 0 in
  F.(
    Variant (dummy_range,
             Datatype.Label "None", (Datatype.Type "option", [TyForall (alpha, TyVar alpha)]), unit)
  )

let none_pattern =
  let alpha = 0 in
  F.(
    PVariant (dummy_range,
      Datatype.Label "None",
      (Datatype.Type "option", [TyForall (alpha, TyVar alpha)]),
      unit_pattern
    )
  )

(* Some Λα.λx:α.x *)
let some =
  let alpha = 0 in
  F.(
    Variant (dummy_range,
      Datatype.Label "Some",
      (Datatype.Type "option", [ TyForall (alpha, TyArrow (TyVar alpha, TyVar alpha)) ]),
      TyAbs (dummy_range, alpha, Abs (dummy_range, "x", TyVar alpha, Var (dummy_range, "x")))
    )
  )

let some_pattern =
  let alpha = 0 in
  F.(
    PVariant (dummy_range,
      Datatype.Label "Some",
      (Datatype.Type "option", [ TyForall (alpha, TyVar alpha) ]),
      PWildcard dummy_range
    )
  )

(* Λα. match None with
       | None     -> ()
       | Some (_) -> ()
 *)
let match_none = F.(
  let alpha = 0 in
  TyAbs(dummy_range, alpha,
    Match (dummy_range,
      TyProduct [] ,
      none,
      [
        (none_pattern , unit)
      ;
        (some_pattern , unit)
      ]
      )
    )
  )

let test_ok_from_ast datatype_env t () =
  Alcotest.(check unit) "type inference" (Test.CheckF.test datatype_env t) ()

let test_case msg datatype_env t =
  Alcotest.(test_case msg `Quick (test_ok_from_ast datatype_env t))

let test_suite =
  [(
    "test F ast",
    [ test_case "let prod" Datatype.Env.empty let_prod;
      test_case "match with prod" Datatype.Env.empty match_with_prod;
      test_case "unit" option_env unit;
      test_case "none" option_env none;
      test_case "some" option_env some;
      test_case "match none" option_env match_none;
    ]
  )]

let () =
  Alcotest.run "F test suite" test_suite
